function [sympoles,residues] = aaa_real(omega,f,tol,m_min);
% [sympoles,residues] = aaa_real(omega,f,tol,m_min);
%
% version of the AAA algorithm for real rational approximation of an
% analytic function with a real Laurent expansion at infinity;
% data must be complex conjugate and nonreal(!)
%
% input:  omega,f          abscissa and function values on unit circle
%                          number of data points must be even, and neither
%                          1 nor -1 belong to omega; flipped data are complex
%                          conjugates of the original data, i.e.,
%                            omega(length(omega)+1-k) = conj(omega(k))
%         tol              tolerance for AAA algorithm (default: 1e-13)
%         m_min            minimal number of poles
%
% output: sympoles        (compl. conj.) poles of the rational approximation
%         residues        (compl. conj.) weights/residuals associated 
%                           with the poles
%
% written by Martin Hanke (29.09.2022)
% based on code published in Nakatsukasa, Sete, Trefethen (2018)

global output_file

fprintf(output_file,'AAA: number of prescribed poles: %1i\n',m_min);
omega = omega(:);    % make sure, omega and f are column vectors
f = f(:);
f_remain = f;

norm_f = norm(f,inf);

n = length(f_remain);
nhalf = round(n/2);

if m_min > nhalf-2
   fprintf(output_file,'*** number of requested poles too large\n');
   error('*** number of requested poles too large')
end

f_remain = f_remain(2:nhalf);
z_remain = omega;
z_remain(nhalf+1) = [];
z_remain(1) = [];

% initialization

z_interp = [omega(1);omega(nhalf+1)];
f_interp = [f(1);f(nhalf+1)];

C = [1./(z_remain-z_interp(1)),1./(z_remain-z_interp(2))];  

F = diag([f_remain;flipud(conj(f_remain))]);
F_interp = diag(f_interp);
A = F*C - C*F_interp;             % Loewner matrix

[U,S,V] = svd(A);

w = V(:,end);

% compute singular vector with complex conjugate entries:
w1 = real(w);
w1 = w1/norm(w1);
w = w1;

m = length(w);

N = C*(w.*f_interp);        % numerator at all remaining grid points
D = C*w;                    % denominator at all remaining grid points
r = N./D;                   % rational function at all remaining grid points
l = length(r);

lhalf = round(l/2);

[err,ind] = max(abs(f_remain-r(1:lhalf))); % max error at remaining grid points
%errs = err;

% AAA greedy iteration

while (err>tol*norm_f) | (m<=m_min),

   z0 = z_remain(ind);

   z_interp = [z_interp;z0;conj(z0)];
   f_interp = [f_interp;f_remain(ind);conj(f_remain(ind))];

   f_remain(ind) = [];  

   l = length(z_remain);
   z_remain([ind,l+1-ind]) = [];

   C([ind,l+1-ind],:) = [];   % delete two rows of Cauchy matrix

   % add two new columns to the Cauchy matrix:
   C = [C,1./(z_remain-z0),1./(z_remain-conj(z0))];  

   F = diag([f_remain;flipud(conj(f_remain))]);
   F_interp = diag(f_interp);
   A = F*C - C*F_interp;     % Loewner matrix

   [U,S,V] = svd(A);

   w = V(:,end);

   % compute singular vector with complex conjugate entries:
   w1 = zeros(size(w));
   w1(1) = w(1) + conj(w(1));
   w1(2) = w(2) + conj(w(2));
   w1(3:2:end) = w(3:2:end)+conj(w(4:2:end));
   w1(4:2:end) = w(4:2:end)+conj(w(3:2:end));
   w1 = w1/norm(w1);

   w = w1;
   m = length(w);

   N = C*(w.*f_interp);   % numerator at all remaining grid points
   D = C*w;               % denominator at all remaining grid points
   r = N./D;              % rational approximation at all remaining grid points
   l = length(r);
   lhalf = round(l/2);

   % max error at remaining grid points:
   [err,ind] = max(abs(f_remain-r(1:lhalf)));  

   if (m+2>nhalf),
      fprintf(output_file,...
              'AAA: half the interpolation constraints were used\n');
      err = 0;   % terminate loop
   end;

end;   

B = eye(m+1);
B(1,1) = 0;
E = [0 w.'; ones(m,1) diag(z_interp)];
poles = eig(E,B);
poles = poles(~isinf(poles));

C = 1./bsxfun(@minus,omega,poles.'); 

rat = @(zz) feval(@rhandle,zz,z_interp,f_interp,w);
dz = 1e-5*exp(2i*pi*(1:4)/4);
a = rat(bsxfun(@plus,poles,dz))*dz.'/4;

sympoles = [];
residues = [];
% exact symmetrization of poles and residues
while length(poles)>0
   [~,l] = min(abs(poles(1:end)-conj(poles(1))));
   if l==1,
      sympoles = [sympoles; real(poles(1))];
      residues = [residues; real(a(1))];
      poles(1) = [];
      a(1) = []; 
   else
      sympoles = [sympoles;(poles(1)+conj(poles(l)))/2;...
                           (conj(poles(1))+poles(l))/2];
      residues = [residues;(a(1)+conj(a(l)))/2;(conj(a(1))+a(l))/2];
      poles([1,l]) = [];
      a([1,l]) = [];
   end;
end;

fprintf(output_file,'AAA: infinity norm of f: %8.6f\n',norm_f);
fprintf(output_file,'AAA: final approximation error: %8.6f\n',err);
fprintf(output_file,'AAA: number of poles: %1i\n\n',length(sympoles));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function rat = rhandle(zz,z,f,w);
zv = zz(:);
CC = 1./bsxfun(@minus,zv,z.');
rat = (CC*(w.*f))./(CC*w);
ii = find(isnan(rat));
for j=1:length(ii),
  rat(ii(j)) = f(find(zv(ii(j))==z));
end;
rat = reshape(rat,size(zz));
