function g = solve_sing_Lure_eq(A,k);
% g = solve_sing_Lure_eq(A,k);
%
% input:  A        mxm matrix with eigenvalues in the left half plane
%                  taken below to be of the form 
%                       A = [-a0,c0';-b0,A0]    
%                  with a0\in\R, b0,c0\in\R^{m-1} and A0\in\R^{(m-1)x(m-1)}
%         k        known number of (still) needed recursive steps
%                  when k>0 then the Lur'e equations are singular
%
% output: g        m-dimensional vector (factor for Sylvester equation)
%   
% the output g admits a solution of the Sylvester equation
%
%    A\Sigma + \Sigma A' = -g*g',
%
% written by Martin Hanke (29.09.2022)

global output_file

m = size(A,1);

a0 = -A(1,1);
if k>0,
   a0 = 0;    % a0 could be nonzero due to round-off
end;
b0 = -A(2:m,1);
c0 = A(1,2:m)';
A0 = A(2:m,2:m);

if a0<0
   fprintf(output_file,'Error: Lure equations are not solvable\n')
   error('*** Lure equations are not solvable: A(1,1)>0\n\n')
end;

if a0>0
   B0 = 2*a0*A0-b0*c0';
   Sigma0 = Riccati(B0',-c0*c0',b0*b0');
   % get file Riccati.m, e.g., from
   % https://de.mathworks.com/matlabcentral/fileexchange/36263-algebraic-riccati-equation-solver
   check = norm(B0*Sigma0+Sigma0*B0'+Sigma0*c0*c0'*Sigma0+b0*b0')/...
           norm(B0*Sigma0);
   fprintf(output_file,'Riccati check = %8.6e (relative)\n',check);
   if check > 1e-6
      fprintf(output_file,'*** Error: Riccati equation not satisfied\n');
      error('*** Riccati equation not satisfied')
   end;
   omega0 = sqrt(2*a0);
   g0 = (b0-Sigma0*c0)/omega0;
else  
   % compute transformation V (see "readme.pdf")

   sigma02 = c0'*b0;
   sigma0 = sqrt(sigma02);
   norm_b0 = norm(b0);
   e1 = zeros(m-1,1);
   e1(1) = 1;
   u = b0/norm_b0 + e1;
   P = eye(m-1) - 2/(norm(u)^2)*u*u';
   V = P + e1*(c0+b0/norm_b0)';

   % V satifies V*b0=sigma02*e_1 and V^{-T}*c0 = e_1

   A1 = V*A0*inv(V);

   if (nargin==1)
      g0 = sigma0 * inv(V) * solve_sing_Lure_eq(A1);
   else
      g0 = sigma0 * inv(V) * solve_sing_Lure_eq(A1,k-1);
   end;

   omega0 = 0;
end;

g = [omega0;g0];
