% algorithm described in the paper
%
% Stochastic modeling of stationary scalar processes in continuous time
% from autocorrelation data
%
% by Martin Hanke
%
% Matlab source copyright by Martin Hanke (29.09.2022)
%
% see documentation in accompanying file "readme.pdf"

clear

global output_file   % diagnose output

output_file = fopen('stochmodel_aaa.out','w');

%%%%%% 1. Input of data/Setting of parameter k_* (k_star)

% needs to define:
% n (2n=number of data points)
% t = array of equidistant time data; t(1) = 0; tau = t(2)
% y = array of autocorrelation data at t grid
% tt = array of all available time points 
%   (set tt=t if no further data are known)
% yy = array of autocorrelation data at tt grid
% k_star = 0 or 1, depending on application

% choose between 'Example 5.1', 'Example 5.2', and 'Example 5.3' or provide
% own data

example = 'Example 5.1';

switch example
  case 'Example 5.1',
     n = 25;
     [t,y,tt,yy,tau] = load_jung_data(n,50);
     k_star = 1;
     minimum_poles = 0;
     tol = 1e-3;    % tolerance for AAA data fit
     fprintf('\nstochmodel_aaa: Example 5.1\n\n');
     fprintf(output_file, 'stochmodel_aaa: Example 5.1\n\n');
  case 'Example 5.2',
     n = 10;
     [t,y,tt,yy,tau] = load_jung_data(n,60);
     k_star = 1;
     minimum_poles = 0;
     tol = 1e-3;    % tolerance for AAA data fit
     fprintf('\nstochmodel_aaa: Example 5.2\n\n');
     fprintf(output_file, 'stochmodel_aaa: Example 5.2\n\n');
  case 'Example 5.3',
     n = 23;   % select n=23, n=24, or n=25 to reproduce results
     [t,y,tt,yy,tau] = load_shea_data(n,6);
     k_star = 1;
     minimum_poles = 0;
     tol = 1e-3;    % tolerance for AAA data fit
     fprintf('\nstochmodel_aaa: Example 5.3\n\n');
     fprintf(output_file, 'stochmodel_aaa: Example 5.3\n\n');
end;

n2 = 2*n;
sigma = sqrt(y(1));
sigma2 = y(1);

fprintf(output_file,'number of data points: 2n=%1i\n',n2);
fprintf(output_file,'time step width: tau = %.1e\n',tau);
if k_star == 0,
   fprintf(output_file,'preassigned constraint: (4.9)\n\n');
elseif k_star == 1,
   fprintf(output_file,'preassigned constraints: (4.9) and (4.10)\n\n'); 
else
   fprintf(output_file,'*** Error: k_star must be zero or one\n');
   error('*** k_star must be zero or one')
end;

%%%%%% 2. Application of the AAA algorithm

e_last = zeros(n2,1);
e_last(n2) = 1;
omega = fft(e_last);  
f = fft(y)./omega;
norm_f = norm(f,inf);

[sympoles,residues] = aaa_real(omega,f,tol,minimum_poles);

% eliminate spurious poles
res_tol = 1e-3*max(abs(y));

inds = abs(residues)<res_tol;
sympoles(inds) = [];
residues(inds) = [];

fprintf(output_file,'%1i spurious poles discarded\n\n',sum(inds));

% check for divergent poles
outside_poles = find(abs(sympoles)>=1);
if max(abs(sympoles))>=1

   outside_pole_residues = abs(residues(outside_poles));
   fprintf(output_file,'*** Warning: significant pole(s) in the exterior');
   fprintf(output_file,' of the unit disk\n');
   fprintf(output_file,...
           '             corresponding absolute relative residues:\n');
   fprintf(output_file,'  %8.6f\n',outside_pole_residues/y(1));
   fprintf(output_file,'    Better check the data fit of the given samples below\n\n');

   sympoles(outside_poles) = [];
   residues(outside_poles) = [];
end;


%%%%%% 3. Determine exponents of Prony series

logpoles = log(sympoles);
neg_inds = find(sympoles<0);
logpoles = [logpoles;logpoles(neg_inds)-2i*pi];

fprintf(output_file,'%1i negative poles duplicated\n\n',length(neg_inds));

k_star = max(k_star,length(neg_inds));  % for a strictly convex problem

lambda = logpoles/tau;
m = length(lambda);

% diagnose output:
lambda_output = zeros(2*length(lambda),1);
lambda_output(1:2:end) = real(lambda);
lambda_output(2:2:end) = imag(lambda);
fprintf(output_file,'exponents of Prony series:\n');
fprintf(output_file,'  %7.4f%+7.4fi\n',lambda_output);
fprintf(output_file,'\n');


%%%%%% 4. Solve constrained minimization problem for the weights

% B constraint matrix; 

fprintf(output_file,'initial value of k_star: %1i\n\n',k_star);

B = ones(1,m);

if k_star>0,
   k = [0:k_star-1]';
   B = [B;(((-1).^k)*ones(size(lambda'))).*...
          (ones(size(k))*(lambda.')).^((2*k+1)*ones(size(lambda')))];
end;

sigma2e1 = sigma2*[1;zeros(k_star,1)];

% least squares system Ga=y
G = exp(t*lambda.');

not_yet_found = 1;

while (size(B,1)<=m) & not_yet_found

   % current inequality constraint:
   B_star = (-1)^(k_star+1)*[(lambda.^(2*k_star+1)).'];

   a = constr_lsq(G,y,B,sigma2e1);

   if (real(B_star*a)>0)
      not_yet_found = 0;
      fprintf(output_file,'constrained minimizer found for k_* = %1i;',k_star);
      fprintf(output_file,' beta_{k_*} = %6e\n\n',real(B_star*a));
   else
      k_star = k_star+1;
      B = [B;B_star];
      sigma2e1 = [sigma2e1;0];
   end;

end;    % loop over k_star

if not_yet_found
   fprintf(output_file,'*** Error: no constrained minimizer found (failure)\n');
   error('*** no constrained minimizer was found')
end;

% diagnose output:
residual = y-G*a;
res_2norm = norm(y-G*a)/norm(y);
inds = find(y==0);
y(inds) = [];
residual(inds) = [];
abs_maxnorm = norm(residual,'inf');
res_maxnorm = norm(residual./y,'inf');
fprintf(output_file,'data fit of the given samples:\n');
fprintf(output_file,'   relative Euclidean norm: %6e\n',res_2norm);
fprintf(output_file,'   absloute maximum norm: %6e\n',abs_maxnorm);
fprintf(output_file,'   relative maximum norm (pointwise): %6e\n\n',...
        res_maxnorm);
a_output = zeros(2*length(a),1);
a_output(1:2:end) = real(a);
a_output(2:2:end) = imag(a);
fprintf(output_file,'weights of Prony series:\n');
fprintf(output_file,'  %7.4f%+7.4fi\n',a_output);
fprintf(output_file,'\n');

phi = exp(tt*lambda.')*a;


%%%%%% 5. Check positivity of phihat (to be sure that the model is passive)

% see "readme.pdf"

numerator = 0;
denominator = 1;

[~,inds] = sort(abs(imag(lambda)));

nu = 1;

% entries corresponding to real exponents first

while isreal(lambda(inds(nu))),

   denominator_nu = [1,lambda(inds(nu))^2];
   if numerator == 0,
      numerator = real(a(inds(nu)))*lambda(inds(nu));
   else
      numerator = conv(numerator,denominator_nu) + ...
                  real(a(inds(nu)))*lambda(inds(nu))*denominator;
   end;
   denominator = conv(denominator,denominator_nu);
   nu = nu+1;
   if nu>m
      break
   end;
end;

% entries corresponding to complex exponents next

while nu<=m,
   if imag(lambda(inds(nu)))>0,
      denominator_nu = [1,2*real(lambda(inds(nu))^2),abs(lambda(inds(nu)))^4];
      numerator_nu = [2*real(a(inds(nu))*lambda(inds(nu))),...
         2*abs(lambda(inds(nu)))^2*real(a(inds(nu))*conj(lambda(inds(nu))))];
      numerator = conv(numerator,denominator_nu) ...
                  + conv(denominator,numerator_nu);
      denominator = conv(denominator,denominator_nu);
   end;
   nu = nu+1;
end;

% leading entries of numerator must be zero because of equality constraint

numerator = numerator(k_star+1:end);

roots_phihat = sqrt(roots(numerator));

% diagnose output:
roots_output = zeros(2*length(roots_phihat),1);
roots_output(1:2:end) = real(roots_phihat);
roots_output(2:2:end) = imag(roots_phihat);
roots_output = [roots_output;-roots_output];
fprintf(output_file,'zeros of phihat:\n');
fprintf(output_file,'  %7.4f%+7.4fi\n',roots_output);

inds = zeros(size(roots_phihat));
for nu=1:length(inds),
   inds(nu) = isreal(roots_phihat(nu));
end;

if sum(inds) == 0,
   fprintf(output_file,'phihat has no real zeros\n\n');
else
   fprintf(output_file,'*** Error: ');
   fprintf(output_file,'phihat has %1i positive zeros (failure)\n',sum(inds));
   error('*** %1i positive zeros of phihat have been found!',sum(inds))
end;


%%%%%% 6. Generation of the preliminary system coefficients J, v, and w

nu=1;
J = zeros(m);
v = zeros(m,1);
w = v;

% entries corresponding to the real exponents first

[~,inds] = sort(abs(imag(lambda)));

while isreal(lambda(inds(nu))),

   J(nu,nu) = lambda(inds(nu));
   v(nu) = 1;
   w(nu) = real(a(inds(nu)));
   nu = nu+1;
   if nu>m
      break
   end;

end;

% entries corresponding to the complex exponents next

while nu<=m,

   J(nu,nu) = real(lambda(inds(nu)));
   J(nu+1,nu+1) = J(nu,nu);
   J(nu,nu+1) = imag(lambda(inds(nu)));
   J(nu+1,nu) = -J(nu,nu+1);
   v(nu) = 1;
   v(nu+1) = 0;
   w(nu) = 2*real(a(inds(nu)));
   w(nu+1) = -2*imag(a(inds(nu)));
   nu = nu+2;

end;


%%%%%% 7. Variable transformation to replace v, w, and J by e_1 and A

% different from Bai/Freund; see "readme.pdf"

norm_w = norm(w);
e1 = zeros(m,1);
e1(1) = 1;

u = w/norm_w + e1;
P = eye(m) - 2/(norm(u)^2)*u*u';
V = P + e1*(v+w/norm_w)';

% V satifies Vw=sigma2*e_1 and V^{-T}v = e_1

A = V*J*inv(V);


%%%%%% 8. Solve (singular) Lur'e equation

% solve_sing_Lure_eq.m only returns the vector g
% it could also return the covariance matrix Sigma.
% Here Sigma is computed instead from Sylvester's equation as a double check

g = sigma * solve_sing_Lure_eq(A,k_star);

Sigma = sylvester(A,A',-g*g');

check = norm(A*Sigma+Sigma*A'+g*g')/norm(A*Sigma);
fprintf(output_file,'norm(A*Sigma+Sigma*A''+g*g'') = %8.6e (relative)\n',check);
if check > 1e-9
   fprintf(output_file,'*** Error: Sylvester equation not satisfied\n');
   error('*** Sylvester equation not satisfied')
end;

check = norm(Sigma-Sigma')/norm(Sigma);
fprintf(output_file,'norm(Sigma-Sigma'') = %8.6e\n',check);
if check > 1e-9
   fprintf(output_file,'*** Error: Sigma fails to be symmetric\n');
   error('*** Sigma fails to be symmetric')
end;

check = min(eig(Sigma));
fprintf(output_file,'smallest eigenvalue of Sigma: %8.6e\n',check);
if check < 0
   fprintf(output_file,'*** Error: Sigma has negative eigenvalues\n');
   error('*** Sigma has negative eigenvalues')
end;

check = norm(Sigma(:,1)-sigma2*e1)/sigma2;
fprintf(output_file,'norm(Sigma(:,1)-sigma2*e1) = %8.6e (relative)\n',check);
if check > 1e-9
   fprintf(output_file,'*** Error: ');
   fprintf(output_file,'First column of Sigma is different from sigma^2*e1\n');
   error('*** First column of Sigma is different from sigma^2*e1')
end;

fprintf(output_file,...
        '\nsuccessful computation of the approximating Markov model\n');

fclose(output_file);

fprintf('successful computation of the approximating Markov model\n\n');

